/**
 * Greek Named Entity Recognizer.
 * Copyright (C) 2005  G. Lucarelli
 *
 * This file is part of Greek Named Entity Recognizer.
 *
 * Sentence Splitter is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Sentence Splitter is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Foobar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package ipl.nerc;

import java.io.*;
import java.util.StringTokenizer;
import java.util.Arrays;

import libsvm.svm;
import libsvm.svm_node;
import libsvm.svm_model;

import ipl.nerc.features.PersonFeaturesManager;
import ipl.nerc.regex.timex.TimexMatcher;
import ipl.nerc.features.FeaturesManager;
import ipl.nerc.lists.ListManager;
import ipl.splitter.SentenceSplitter;
import ipl.util.*;


/**
 * <p>Description: Trains a classifier for person names recognition.</p>
 *
 * <p>Copyright: Copyright (c) 2005</p>
 * @version 0.1
 * @see <a href="http://www.csie.ntu.edu.tw/~cjlin/libsvm">libSVM</a>
 */
public class Trainer {

  public Trainer() {
  } // constructor: Trainer

  //////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////// TRAIN ////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////

  public void TRAIN_FP(String dirName, String classifierfn, ListManager listManager,
                       TimexMatcher matcher, SentenceSplitter splitter,
                       double cFP, double gammaFP) throws Exception {
    File curr;
    String vectorsFile = "TRAIN1";
    int numOfTrainingVectors = 0;

    FeaturesFileManager outputFileManager = new FeaturesFileManager(vectorsFile);
    DirectoryManager trainDir = new DirectoryManager(dirName);

    //        train set
    while ((curr = trainDir.getNextFile()) != null) {
      if (curr.isFile()) {
        String attr[];
        int distanceFromStartOfEntity = -1;

        //   
        String text = FileManager.read(curr);
        FeaturesManager fm = new PersonFeaturesManager(text, listManager, matcher, splitter);

        //    tokens  
        while ( (attr = fm.getNextFeatureVector(distanceFromStartOfEntity, null)) != null) {
          if (attr.length != 1) {
            numOfTrainingVectors++;
            outputFileManager.setData(attr);
          } // if
          else if (attr[0].compareTo("1") == 0) {
            System.out.println(curr.getName());
            System.out.println(fm.tokenAt(fm.getOffset() - 1).getToken());
          } // else if
          //       .
          //         
          // .
          if ((attr[0].compareTo("1") == 0) && (distanceFromStartOfEntity == -1)) {
            distanceFromStartOfEntity = 1;
          } // if
          else if (attr[0].compareTo("1") == 0) {
            distanceFromStartOfEntity++;
          } // else if
          else if ((attr[0].compareTo("-1") == 0) && (distanceFromStartOfEntity != -1)) {
            distanceFromStartOfEntity = -1;
          } // else if
        } // while
      } // if
    } // while
    outputFileManager.close();

    FileInputStream fis = new FileInputStream(vectorsFile);
    byte buff[] = new byte[2048];
    fis.read(buff);
    fis.close();
    String temp = new String(buff);
    temp = temp.substring(0, temp.indexOf("\n"));
    StringTokenizer strtok = new StringTokenizer(temp, " ");
    int max_index = strtok.countTokens() - 1;

    System.out.println("Training vectors (first pass): " + numOfTrainingVectors);

    System.out.println("training...");
    NativeTrainer.train(vectorsFile.toCharArray(), vectorsFile.toCharArray().length,
                        classifierfn.toCharArray(), classifierfn.toCharArray().length,
                        max_index, true, cFP, gammaFP);
    System.out.println("DONE");

    (new File(vectorsFile)).deleteOnExit();
  } // TRAIN_FP

  //////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////

  public void TRAIN_SP(String dirName, String inputModelFileName, String outputModelFileName,
                       ListManager firstPassListManager, ListManager secondPassListManager,
                       TimexMatcher matcher, SentenceSplitter splitter,
                       double cSP, double gammaSP) throws Exception {
    double margin = 0.8;
    File curr;
    String vectorsFile = "TRAIN2";
    int numOfTrainingVectors = 0;

    DirectoryManager trainDir = new DirectoryManager(dirName);
    FeaturesFileManager outputFileManager = new FeaturesFileManager(vectorsFile);

    svm_model model = svm.svm_load_model(inputModelFileName);

    int nr_class = svm.svm_get_nr_class(model);
    int []categories = new int[nr_class];
    svm.svm_get_labels(model, categories);
    double[] prob_estimates = new double[nr_class];

    //        train set
    while ( (curr = trainDir.getNextFile()) != null) {
      if (curr.isFile()) {

        ///////////////////////////////////////////////
        //       //
        ///////////////////////////////////////////////

        String attr[];
        int distanceFromStartOfEntity = -1;
        String temp[] = new String[256];
        int listNum = 0;

        //   
        String text = FileManager.read(curr);
        FeaturesManager vectorsFP = new PersonFeaturesManager(text, firstPassListManager, matcher, splitter);

        //    tokens  
        while ( (attr = vectorsFP.getNextFeatureVector(distanceFromStartOfEntity, null)) != null) {
          double answer, probability;
          if (attr.length != 1) {
            int numOfAttributes = attr.length - 1;
            svm_node node[] = new svm_node[numOfAttributes];
            for (int i = 1; i <= numOfAttributes; i++) {
              node[i - 1] = new svm_node();
              node[i - 1].index = i;
              node[i - 1].value = Auxiliary.atof(attr[i]);
            } // for
            // PREDICT
            answer = svm.svm_predict_probability(model, node, prob_estimates);
            if (categories[0] == -1) {
              probability = prob_estimates[1] - prob_estimates[0];
            } // if
            else {
              probability = prob_estimates[0] - prob_estimates[1];
            } // else
          } // if
          else {
            answer = -1;
            probability = -1;
          } // else
          vectorsFP.tokenAt(vectorsFP.getOffset() - 1).setPredictedProbability(probability);

          //       .
          //         
          // .
          if ((answer == 1) && (distanceFromStartOfEntity == -1)) {
            distanceFromStartOfEntity = 1;
            String currTok = vectorsFP.tokenAt(vectorsFP.getOffset() - 1).getToken();
            if ((currTok.length() >= 3) && (probability > margin)) {
              temp[listNum++] = currTok.toLowerCase();
            } // if
          } // if
          else if (answer == 1) {
            distanceFromStartOfEntity++;
            String currTok = vectorsFP.tokenAt(vectorsFP.getOffset() - 1).getToken();
            if ((currTok.length() >= 3) && (probability > margin)) {
              temp[listNum++] = currTok.toLowerCase();
            } // if
          } // else if
          else if ((answer == -1) && (distanceFromStartOfEntity != -1)) {
            distanceFromStartOfEntity = -1;
          } // else if
        } // while

        //      
        String []list = Auxiliary.getSubList(temp, listNum);
        Arrays.sort(list);

        ///////////////////////////////////////////////////////////////
        //         //
        ///////////////////////////////////////////////////////////////

        FeaturesManager vectorsSP = new PersonFeaturesManager(vectorsFP.getTokens(), secondPassListManager);

        distanceFromStartOfEntity = -1;

        //    tokens  
        while ( (attr = vectorsSP.getNextFeatureVector(distanceFromStartOfEntity, list)) != null) {
          if (attr.length != 1) {
            numOfTrainingVectors++;
            outputFileManager.setData(attr);
          } // if
          //       .
          //         
          // .
          if ((attr[0].compareTo("1") == 0) && (distanceFromStartOfEntity == -1)) {
            distanceFromStartOfEntity = 1;
          } // if
          else if (attr[0].compareTo("1") == 0) {
            distanceFromStartOfEntity++;
          } // else if
          else if ((attr[0].compareTo("-1") == 0) && (distanceFromStartOfEntity != -1)) {
            distanceFromStartOfEntity = -1;
          } // else if
        } // while

        vectorsFP = null;
        vectorsSP = null;
        list = null;
        temp = null;
      } // if
    } // while
    outputFileManager.close();

    //   
    FileInputStream fis = new FileInputStream(vectorsFile);
    byte buff[] = new byte[2048];
    fis.read(buff);
    fis.close();
    String temp = new String(buff);
    temp = temp.substring(0, temp.indexOf("\n"));
    StringTokenizer strtok = new StringTokenizer(temp, " ");
    int max_index = strtok.countTokens() - 1;

    System.out.println("Training vectors (second pass): " + numOfTrainingVectors);

    System.out.println("training...");
    NativeTrainer.train(vectorsFile.toCharArray(), vectorsFile.toCharArray().length,
                        outputModelFileName.toCharArray(), outputModelFileName.toCharArray().length,
                        max_index, true, cSP, gammaSP);
    System.out.println("DONE");

    (new File(vectorsFile)).deleteOnExit();
  } // TRAIN_SP

  //////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////

  private static void printParametersAndExit(String str) {
    System.err.println(str);
    System.err.println();
    System.err.println();
    System.err.println("java ipl.nerc.Trainer -1 <classifier1> [-2 <classifier2>] -f <train1> [-s <train2>] [options]");
    System.err.println();
    System.err.println("-1");
    System.err.println("\tThe first pass classifier output filename");
    System.err.println("-2");
    System.err.println("\tThe second pass classifier output filename");
    System.err.println("-f");
    System.err.println("\tThe first pass train directory path");
    System.err.println("-s");
    System.err.println("\tThe second pass train directory path");
    System.err.println("\nOPTIONS");
    System.err.println("=======");
    System.err.println("-m");
    System.err.println("\tMode:");
    System.err.println("\t\t1. Train the first pass SVM only");
    System.err.println("\t\t2. Train both the first and second pass SVMs\t(default)");
    System.err.println("-l");
    System.err.println("\tThe lists directory\t\t\tdefault \".\\lists\"");
    System.err.println("-t");
    System.err.println("\tThe TIMEX patterns filename\t\tdefault \".\\TIMEX_PATTERNS\"");
    System.err.println("-p");
    System.err.println("\tThe sentence splitter filename\t\tdefault \".\\SENTENCE_SPLITTER\"");
    System.err.println("-a");
    System.err.println("\tThe parameter c of first SVM\t\tdefault 16");
    System.err.println("-b");
    System.err.println("\tThe parameter gamma of first SVM\tdefault 0.125");
    System.err.println("-c");
    System.err.println("\tThe parameter c of second SVM\t\tdefault 16");
    System.err.println("-d");
    System.err.println("\tThe parameter gamma of second SVM\tdefault 0.125");
    System.err.println();
    System.exit(1);
  } // printParametersAndExit

  public static void main(String args[]) throws Exception {
    if ((args.length < 4) || (args.length % 2 == 1)) {
      printParametersAndExit("Invalid parameters number!");
    } // if

    int mode = 2;
    String class1 = "";
    String class2 = "";
    String tdata1 = "";
    String tdata2 = "";
    String listsdir = "./lists";
    String patternsfn = "./TIMEX_PATTERNS";
    String splitterfn = "./SENTENCE_SPLITTER";
    double cFP = 16.0;//a
    double gFP = 0.125;//b
    double cSP = 16.0;//c
    double gSP = 0.125;//d
    for (int i = 0; i < args.length; i += 2) {
      if (args[i].length() != 2) {
        printParametersAndExit("Invalid parameters!");
      } // if
      if (args[i].charAt(0) != '-') {
        printParametersAndExit("Invalid parameters!");
      } // if
      char param = args[i].charAt(1);
      switch (param) {
        case '1':
          class1 = args[i + 1];
          break;
        case '2':
          class2 = args[i + 1];
          break;
        case 'f':
          tdata1 = args[i + 1];
          if (!(new File(tdata1)).isDirectory()) {
            printParametersAndExit("Invalid first pass train directory path!");
          } // if
          break;
        case 's':
          tdata2 = args[i + 1];
          if (!(new File(tdata2)).isDirectory()) {
            printParametersAndExit("Invalid second pass train directory path!");
          } // if
          break;
        case 'l':
          listsdir = args[i + 1];
          break;
        case 't':
          patternsfn = args[i + 1];
          break;
        case 'p':
          splitterfn = args[i + 1];
          break;
        case 'm':
          mode = Integer.parseInt(args[i + 1]);
          if ((mode != 1) && (mode != 2)) {
            printParametersAndExit("Invalid mode!");
          } // if
          break;
        case 'a':
          cFP = Double.parseDouble(args[i + 1]);
          break;
        case 'b':
          gFP = Double.parseDouble(args[i + 1]);
          break;
        case 'c':
          cSP = Double.parseDouble(args[i + 1]);
          break;
        case 'd':
          gSP = Double.parseDouble(args[i + 1]);
          break;
        default:
          printParametersAndExit("Invalid parameters!");
          break;
      } // switch
    } // for

    if (class1.compareTo("") == 0) {
      printParametersAndExit("You must give the first pass classifier output filename!");
    } // if
    if (tdata1.compareTo("") == 0) {
      printParametersAndExit("You must give the first pass train directory path!");
    } // if
    if (mode == 2) {
      if (class2.compareTo("") == 0) {
        printParametersAndExit("You must give the second pass classifier output filename!");
      } // if
      if (tdata2.compareTo("") == 0) {
        printParametersAndExit("You must give the second pass train directory path!");
      } // if
    } // if

    if (!(new File(listsdir)).exists() || !(new File(listsdir)).isDirectory()) {
      printParametersAndExit("Invalid lists directory!");
    } // if
    if (!(new File(patternsfn)).exists() || !(new File(patternsfn)).isFile()) {
      printParametersAndExit("Invalid TIMEX patterns filename!");
    } // if
    if (!(new File(splitterfn)).exists() || !(new File(patternsfn)).isFile()) {
      printParametersAndExit("Invalid sentence splitter filename!");
    } // if

    TimexMatcher matcher = new TimexMatcher(patternsfn);
    SentenceSplitter splitter = new SentenceSplitter(splitterfn);
    ListManager listManagerFP = ListManager.initListManager(tdata1, "FP", splitter, listsdir);

    Trainer trainer = new Trainer();

    trainer.TRAIN_FP(tdata1, class1, listManagerFP, matcher, splitter, cFP, gFP);
    if (mode == 2) {
      ListManager listManagerSP = ListManager.initListManager(tdata2, "SP", splitter, listsdir);
      trainer.TRAIN_SP(tdata2, class1, class2, listManagerFP, listManagerSP, matcher, splitter, cSP, gSP);
    } // else
  } // end of main

} // end of class Trainer
