/**
 * Greek Named Entity Recognizer.
 * Copyright (C) 2005  G. Lucarelli
 *
 * This file is part of Greek Named Entity Recognizer.
 *
 * Sentence Splitter is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Sentence Splitter is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Foobar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package ipl.nerc.features;

import java.util.regex.*;
import java.util.Hashtable;
import java.util.Arrays;

import ipl.core.*;
import ipl.nerc.regex.timex.TimexMatcher;
import ipl.splitter.SentenceSplitter;
import ipl.nerc.lists.*;

/**
 * <p>Description: Abstract class. Retrieves the feature vectors.</p>
 *
 * <p>Copyright (c) 2005</p>
 * @version 0.1
 */
public abstract class FeaturesManager {

  protected Tokenizer tokens;
  protected int offset;
  protected ListManager lists;

  public FeaturesManager(String text, ListManager lists,
                                 TimexMatcher matcher,
                                 SentenceSplitter splitter) {
    this.tokens = new Tokenizer(text, matcher, splitter);
    this.offset = 0;
    this.lists = lists;
  } // constructor: FeaturesManager

  public FeaturesManager(Tokenizer tokens, ListManager lists) {
    this.tokens = tokens;
    this.offset = 0;
    this.lists = lists;
  } // constructor: FeaturesManager

  public void reset() {
    this.offset = 0;
  } // reset

  public TokenNode tokenAt(int pos) {
    return tokens.get(pos);
  } // tokenAt

  public int getOffset() {
    return offset;
  } // getOffset

  public Tokenizer getTokens() {
    return tokens;
  } // getTokens

  private boolean insidePeriod(int pos) {
    if ( (pos < 0) || (pos >= this.tokens.size())) {
      return false;
    } // if

    if (pos < this.offset) {
      for (int i = pos; i < this.offset; i++) {
        if (tokens.get(i).getEndOfPeriod()) {
          return false;
        } // if
      } // for
    } // if
    else if (pos > this.offset) {
      for (int i = this.offset; i <= pos; i++) {
        if (tokens.get(i).getEndOfPeriod()) {
          return false;
        } // if
      } // for
    } // else if

    return true;
  } // insidePeriod

  protected String isGrWord(int pos) {
    if (!insidePeriod(pos)) {
      return "-1";
    } // if

    String grWord = "\\p{InGreek}*";
    String tok = tokens.get(pos).getToken();

    if (Pattern.matches(grWord, tok)) {
      return "1";
    } // if
    else {
      return "-1";
    } // else

  } // isGrWord

  protected String isEnWord(int pos) {
    if (!insidePeriod(pos)) {
      return "-1";
    } // if

    String enWord = "[a-zA-Z]*";
    String tok = tokens.get(pos).getToken();

    if (Pattern.matches(enWord, tok)) {
      return "1";
    } // else if
    else {
      return "-1";
    } // else

  } // isEnWord

  protected String isNumber(int pos) {
    if (!insidePeriod(pos)) {
      return "-1";
    } // if

    String number = "[0-9]*";
    String tok = tokens.get(pos).getToken();

    if (Pattern.matches(number, tok)) {
      return "1";
    } // if
    else {
      return "-1";
    } // else

  } // isNumber

  protected String isDot(int pos) {
    if (!insidePeriod(pos)) {
      return "-1";
    } // if

    String tok = tokens.get(pos).getToken();

    if (tok.compareTo(".") == 0) {
      return "1";
    } // if
    else {
      return "-1";
    } // else

  } // isDot

  protected String isComma(int pos) {
    if (!insidePeriod(pos)) {
      return "-1";
    } // if

    String tok = tokens.get(pos).getToken();

    if (tok.compareTo(",") == 0) {
      return "1";
    } // if
    else {
      return "-1";
    } // else

  } // isComma

  protected String isSlash(int pos) {
    if (!insidePeriod(pos)) {
      return "-1";
    } // if

    String tok = tokens.get(pos).getToken();

    if (tok.compareTo("/") == 0) {
      return "1";
    } // if
    else {
      return "-1";
    } // else

  } // isSlash

  protected String isDash(int pos) {
    if (!insidePeriod(pos)) {
      return "-1";
    } // if

    String tok = tokens.get(pos).getToken();

    if (tok.compareTo("-") == 0) {
      return "1";
    } // if
    else {
      return "-1";
    } // else

  } // isDash

  protected String isEndOfPeriod(int pos) {
    if ( (pos < 0) || (pos >= this.tokens.size())) {
      return "1";
    } // if
    if (tokens.get(pos).getEndOfPeriod()) {
      return "1";
    } // if

    return "-1";
  } // isEndOfPeriod

  protected String getLengthProperty(int pos) {
    if (!insidePeriod(pos)) {
      return "-1";
    } // if

    String tok = tokens.get(pos).getToken();
    double l = tok.length();

    if (l >= 10) {
      return "1";
    } // if
    else {
      return Double.toString(((l / 10) * 2) - 1);
    } // else
  } // getLengthProperty

  protected String firstCapital(int pos) {
    if (!insidePeriod(pos)) {
      return "-1";
    } // if

    String tok = tokens.get(pos).getToken();

    if ((tok.substring(0, 1).compareTo(tok.substring(0, 1).toUpperCase()) == 0)
        && (tok.substring(1).compareTo(tok.substring(1).toLowerCase()) == 0)) {
      return "1";
    } // if

    return "-1";
  } // firstCapital

  protected String isCapitalized(int pos) {
    if (!insidePeriod(pos)) {
      return "-1";
    } // if

    String tok = tokens.get(pos).getToken();

    if (tok.compareTo(tok.toUpperCase()) == 0) {
      return "1";
    } // if

    int countC = 0;
    int countL = 0;
    String capLetter = "[A-Z-٢]";
    for (int i = 1; i < tok.length(); i++) {
      if (Pattern.matches(capLetter, String.valueOf(tok.charAt(i)))) {
        countC++;
      } // if
      else {
        countL++;
      } // else
    } // for
    if ( (countC > 0) && (countL > 0)) {
      return "1";
    } // if

    return "-1";
  } // isCapitalized

  protected String greekEnding(int pos) {
    if (!insidePeriod(pos)) {
      return "-1";
    } // if

    String tok = tokens.get(pos).getToken();

    if (tok.toLowerCase().endsWith("") || tok.endsWith("") ||
        tok.toLowerCase().endsWith("") || tok.toLowerCase().endsWith("") ||
        tok.toLowerCase().endsWith("") || tok.toLowerCase().endsWith("") ||
        tok.toLowerCase().endsWith("") || tok.toLowerCase().endsWith("") ||
        tok.toLowerCase().endsWith("") || tok.toLowerCase().endsWith("") ||
        tok.toLowerCase().endsWith("") || tok.toLowerCase().endsWith("") ||
        tok.toLowerCase().endsWith("") || tok.toLowerCase().endsWith("") ||
        tok.toLowerCase().endsWith("") || tok.toLowerCase().endsWith("") ||
        tok.toLowerCase().endsWith("") || tok.toLowerCase().endsWith("") ||
        tok.toLowerCase().endsWith("") || tok.toLowerCase().endsWith("") ||
        tok.toLowerCase().endsWith("")) {
      return "1";
    } // if
    else {
      return "-1";
    } // else
  } // greekEnding

  protected String distLessThan(int distFromStart, int level) {
    if (distFromStart == -1) {
      return "-1";
    } // if
    else if (distFromStart <= level) {
      return "1";
    } // else if
    else {
      return "-1";
    } // else
  } // distLessThan

  protected boolean isInside(String categName) {
    if (tokens.get(this.offset).getTagType().compareTo(categName) == 0) {
      return true;
    } // if
    else {
      return false;
    } // else
  } // isInside

  protected String isInList(String[] words, int position) {
    if (!insidePeriod(position)) {
      return "-1";
    } // if

    if (words.length == 0) {
      return "-1";
    } // if

    String temp, token = tokens.get(position).getToken().toLowerCase();
    int pos = Arrays.binarySearch(words, token);
    //   token   
    if (pos >= 0) {
      return "1";
    } // if

    //   token          
    //   .         
    //       token    .
    pos = java.lang.Math.abs(pos) - 1;

    //    
    float res1 = -1;
    if (pos > 0) {
      temp = words[pos - 1];
      int length;
      if (temp.length() < token.length()) {
        length = temp.length();
      } // if
      else {
        length = token.length();
      } // else
      int i = 0;
      boolean end = false;
      while ((i < length) && !end) {
        if (token.charAt(i) != temp.charAt(i)) {
          end = true;
        } // if
        else {
          i++;
        } // else
      } // while
      if (i != 0) {
        res1 = (float) ((2.0 * (float) i / temp.length()) - 1.0);
      } // if
    } // if

    //    
    float res2 = -1;
    if (pos < words.length) {
      temp = words[pos];
      int length;
      if (temp.length() < token.length()) {
        length = temp.length();
      } // if
      else {
        length = token.length();
      } // else
      int i = 0;
      boolean end = false;
      while ((i < length) && !end) {
        if (token.charAt(i) != temp.charAt(i)) {
          end = true;
        } // if
        else {
          i++;
        } // else
      } // while
      if (i != 0) {
        res2 = (float) ((2.0 * (float) i / temp.length()) - 1.0);
      } // if
    } // if

    return Float.toString(Math.max(res1, res2));
  } // isInList

  private int getFrequency(Hashtable<String, ListNode> ht, int p) {
    if (ht.containsKey(tokens.get(p).getToken().toLowerCase())) {
      return ht.get(tokens.get(p).getToken().toLowerCase()).getCount();
    } // if
    return 0;
  } // getFrequency

  protected String isInPrevList(Hashtable<String, ListNode> ht, int level, int dist) {
    int end;
    if (this.offset < dist) {
      end = 0;
    } // if
    else {
      end = this.offset - dist;
    } // else

    int count = 0;
    int i = this.offset - 1;
    while ((i >= end) && (!tokens.get(i).getEndOfPeriod())) {
      count += getFrequency(ht, i);
      i--;
    } // while

    if (count < (level / 3)) {
      return "-1.0";
    } // if
    else if (count < 2 * level / 3) {
      return "-0.75";
    } // else if
    else if (count < 3 * level / 3) {
      return "-0.5";
    } // else if
    else if (count < 4 * level / 3) {
      return "-0.25";
    } // else if
    else if (count < 5 * level / 3) {
      return "0.0";
    } // else if
    else if (count < 6 * level / 3) {
      return "0.25";
    } // else if
    else if (count < 7 * level / 3) {
      return "0.5";
    } // else if
    else if (count < 8 * level / 3) {
      return "0.75";
    } // else if
    else {
      return "1.0";
    } // else
  } // isInPrevList

  protected String insideSubtitle(int pos) {
    if (!insidePeriod(pos)) {
      return "-1";
    } // if

    if (tokens.get(pos).isSubtitle()) {
      return "1";
    } // if
    return "-1";
  } // insideSubtitle

  protected String getTokenFirstPassProbability(int pos) {
    if (!insidePeriod(pos)) {
      return "-1";
    } // if

    return Double.toString(tokens.get(pos).getPredictedProbability());
  } // getTokenFirstPassProbability

  protected String inPlural(int pos) {
    if (!insidePeriod(pos)) {
      return "-1";
    } // if

    String tok = tokens.get(pos).getToken();

    if (tok.endsWith("") || tok.endsWith("") ||
        tok.endsWith("") || tok.endsWith("") ||
        tok.endsWith("") || tok.endsWith("") ||
        tok.endsWith("") || tok.endsWith("")) {
      return "1";
    } // if
    else {
      return "-1";
    } // else
  } // inPlural

  protected String getEndingProperty(int pos) {
      if (!insidePeriod(pos)) {
        return "-1";
      } // if

      String tok = tokens.get(pos).getToken();

      if (tok.toLowerCase().endsWith("") || tok.toLowerCase().endsWith("") ||
          tok.toLowerCase().endsWith("") || tok.toLowerCase().endsWith("") ||
          tok.endsWith("") || tok.endsWith("")) {
        return "1";
      } // else if
      else {
        return "-1";
      } // else
  } // getEndingProperty

  protected String isMasculineNominative(int pos) {
    if (!insidePeriod(pos)) {
      return "-1";
    } // if

    String tok = tokens.get(pos).getToken();

    if (tok.toLowerCase().endsWith("")) {
      return "1";
    } // if
    else if (tok.endsWith("")) {
      return "1";
    } // else if
    else {
      return "-1";
    } // else
  } // isMasculineNominative

  protected String greekNamePrefixProperty(int pos) {
    if (!insidePeriod(pos)) {
      return "-1";
    } // if

    String tok = tokens.get(pos).getToken();

    if (tok.toLowerCase().startsWith("") || tok.toLowerCase().startsWith("")) {
      return "1";
    } // if
    else if (tok.toLowerCase().startsWith("") || tok.toLowerCase().startsWith("")) {
      return "1";
    } // else if
    else if (tok.toLowerCase().startsWith("") || tok.toLowerCase().startsWith("")) {
      return "1";
    } // else if
    else if (tok.toLowerCase().startsWith("") || tok.toLowerCase().startsWith("")) {
      return "1";
    } // else if
    else if (tok.toLowerCase().startsWith("") || tok.toLowerCase().startsWith("")) {
      return "1";
    } // else if
    else if (tok.toLowerCase().startsWith("") || tok.toLowerCase().startsWith("")) {
      return "1";
    } // else if
    else {
      return "-1";
    } // else
  } // greekNamePrefixProperty

  protected String greekNameSuffixProperty(int pos) {
    if (!insidePeriod(pos)) {
      return "-1";
    } // if

    String tok = tokens.get(pos).getToken();

    if (tok.toLowerCase().endsWith("") || tok.endsWith("") || tok.toLowerCase().endsWith("") || tok.toLowerCase().endsWith("") || tok.toLowerCase().endsWith("")) {
      return "1";
    } // if
    else if (tok.toLowerCase().endsWith("") || tok.endsWith("") || tok.toLowerCase().toLowerCase().endsWith("")) {
      return "1";
    } // else if
    else if (tok.toLowerCase().endsWith("") || tok.endsWith("") || tok.toLowerCase().endsWith("") || tok.toLowerCase().endsWith("")) {
      return "1";
    } // else if
    else if (tok.toLowerCase().endsWith("") || tok.toLowerCase().endsWith("")) {
      return "1";
    } // else if
    else if (tok.toLowerCase().endsWith("") || tok.endsWith("") || tok.toLowerCase().endsWith("")) {
      return "1";
    } // else if
    else if (tok.toLowerCase().endsWith("") || tok.endsWith("") || tok.toLowerCase().endsWith("") || tok.toLowerCase().endsWith("")) {
      return "1";
    } // else if
    else {
      return "-1";
    } // else
  } // greekNameSuffixProperty

  protected String isAdjective(int pos) {
    if (!insidePeriod(pos)) {
      return "-1";
    } // if

    String tok = tokens.get(pos).getToken();

    if (tok.toLowerCase().endsWith("") || tok.toLowerCase().endsWith("") ||
        tok.toLowerCase().endsWith("") || tok.toLowerCase().endsWith("") ||
        tok.toLowerCase().endsWith("")) {
      return "1";
    } // if
    else {
      return "-1";
    } // else
  } // isAdjective

  //    .
  protected String isMr() {
    if ( (this.offset == 0) || (this.offset == 1)) {
      return "-1";
    } // if
    if ( (tokens.get(this.offset - 1).getToken().compareTo(".") == 0) &&
        (tokens.get(this.offset - 2).getToken().compareTo("") == 0)) {
      return "1";
    } // if
    else {
      return "-1";
    } // else
  } // isMr

  //     dist tokens   ..  .
  protected String areMessrs(int dist) {
    int end, i;

    // ..
    if (this.offset < dist) {
      end = 0;
    } // if
    else {
      end = this.offset - dist;
    } // else
    i = this.offset - 4;
    while ((i >= end) && (!tokens.get(i).getEndOfPeriod())) {
      for (int j = i; j < this.offset; j++) {
        if ( (tokens.get(j).getToken().compareTo("") == 0) &&
            (tokens.get(j + 1).getToken().compareTo(".") == 0) &&
            (tokens.get(j + 2).getToken().compareTo("") == 0) &&
            (tokens.get(j + 3).getToken().compareTo(".") == 0)) {
          return "1";
        } // if
      } // for
      i--;
    } // while

    // .
    if (this.offset < dist) {
      end = 0;
    } // if
    else {
      end = this.offset - dist;
    } // else
    i = this.offset - 2;
    while ((i >= end) && (!tokens.get(i).getEndOfPeriod())) {
      for (int j = i; j < this.offset; j++) {
        if ( (tokens.get(j).getToken().compareTo("") == 0) &&
            (tokens.get(j + 1).getToken().compareTo(".") == 0)) {
          return "1";
        } // if
      } // for
      i--;
    } // while

    return "-1";
  } // areMessrs

  protected String getCategory(String type) {
    if (tokens.get(this.offset).getTagType().compareTo(type) == 0) {
      return "1";
    } // if
    else {
      return "-1";
    } // else
  } // getCategory

  public abstract String[] getNextFeatureVector(int distFromStart, String []firstPassList);

} // end of class FeaturesManager
