/**
 * Greek Named Entity Recognizer.
 * Copyright (C) 2005  G. Lucarelli
 *
 * This file is part of Greek Named Entity Recognizer.
 *
 * Sentence Splitter is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Sentence Splitter is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Foobar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package ipl.nerc.lists;

import java.io.*;
import java.util.StringTokenizer;
import java.util.Hashtable;
import java.util.Enumeration;

import ipl.splitter.SentenceSplitter;
import ipl.util.*;

/**
 * <p>Description: Manages the lists.</p>
 *
 * <p>Copyright: Copyright (c) 2005</p>
 * @version 0.1
 */
public class ListManager {

  private Hashtable<String, String> safeList;
  private Hashtable<String, ListNode>[] listsOfRepeatedWords;
  private int[] avgOfRepeatedWordsLists;
  private String[] listOfNames;

  public ListManager(String safeListFileName, String[] PreviouslyRepeatedWordsListsFilesNames, String listOfNamesFileName) {
    if (safeListFileName != null) {
      safeList = loadSafeList(new File(safeListFileName));
    } // if

    if (PreviouslyRepeatedWordsListsFilesNames != null) {
      int num = PreviouslyRepeatedWordsListsFilesNames.length;
      listsOfRepeatedWords = new <String, HashNode>Hashtable[num];
      for (int i = 0; i < num; i++) {
        listsOfRepeatedWords[i] = loadFrequencyList(new File(PreviouslyRepeatedWordsListsFilesNames[i]));
      } // for

      avgOfRepeatedWordsLists = new int[num];
      for (int i = 0; i < num; i++) {
        avgOfRepeatedWordsLists[i] = getAvg(listsOfRepeatedWords[i]);
      } // for
    } // if

    if (listOfNamesFileName != null) {
      listOfNames = loadListOfNames(new File(listOfNamesFileName));
    } // if
  } // constructor: ListManager

  private Hashtable<String, String> loadSafeList(File file) {
    Hashtable<String, String> tmp = new Hashtable<String, String>();
    try {
      FileInputStream fis = new FileInputStream(file.getAbsolutePath());
      byte[] buff = new byte[(int) file.length()];
      fis.read(buff);
      fis.close();
      String text = new String(buff);
      text = Auxiliary.removeNoisySpaces(text);
      StringTokenizer strTok = new StringTokenizer(text, "\n");
      while (strTok.hasMoreTokens()) {
        String tok = strTok.nextToken();
        tmp.put(tok, tok);
      } // while
      strTok = null;
    } // try
    catch (Exception e) {
      e.printStackTrace();
    } // catch
    return tmp;
  } // loadSafeList

  private Hashtable<String, ListNode> loadFrequencyList(File file) {
    Hashtable<String, ListNode> tmp = new Hashtable<String, ListNode>();
    try {
      FileInputStream fis = new FileInputStream(file.getAbsolutePath());
      byte[] buff = new byte[(int) file.length()];
      fis.read(buff);
      fis.close();
      String text = new String(buff);
      text = Auxiliary.removeNoisySpaces(text);
      StringTokenizer strTok = new StringTokenizer(text, " \n");
      while (strTok.hasMoreTokens()) {
        String tok = strTok.nextToken();
        tmp.put(tok, new ListNode(tok, Integer.parseInt(strTok.nextToken())));
      } // while
      strTok = null;
    } // try
    catch (Exception e) {
      e.printStackTrace();
    } // catch
    return tmp;
  } // loadFrequencyList

  private int getAvg(Hashtable<String, ListNode> ht) {
    if (ht.size() == 0) {
      return 0;
    } // if
    Enumeration<ListNode> enumeration = ht.elements();
    int count = 0;
    while (enumeration.hasMoreElements()) {
      count += enumeration.nextElement().getCount();
    } // while

    return ((int) count / ht.size());
  } // getAvg

  private String[] loadListOfNames(File file) {
    String []tmp;
    String text = "";
    try {
      FileInputStream fis = new FileInputStream(file.getAbsolutePath());
      byte[] buff = new byte[(int) file.length()];
      fis.read(buff);
      fis.close();
      text = new String(buff);
      text = Auxiliary.removeNoisySpaces(text);
    } // try
    catch (Exception e) {
      e.printStackTrace();
    } // catch
    StringTokenizer strTok = new StringTokenizer(text, "\n");
    int listLength = strTok.countTokens();
    tmp = new String[listLength];
    for (int i = 0; i < listLength; i++) {
      tmp[i] = strTok.nextToken();
    } // for
    strTok = null;
    return tmp;
  } // loadListOfNames

  //////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////

  public Hashtable<String, String> getSafeList() {
    return safeList;
  } // getSafeList

  public Hashtable<String, ListNode> getRepeatedWordsList(int offset) {
    return listsOfRepeatedWords[offset];
  } // getRepeatedWordsList

  public int getAvgOfRepeatedWordsList(int offset) {
    return avgOfRepeatedWordsLists[offset];
  } // getAvgOfRepeatedWordsList

  public String[] getListOfNames() {
    return listOfNames;
  } // getListOfNames

  //////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////

  public static ListManager initListManager(String dir, String FP_SP,
                                            SentenceSplitter splitter,
                                            String outputDirName) {
    ListManager listManager;
    String lists[] = new String[6];
    String tagType = Constants.PERSON;
    lists[0] = ListCreator.getPreviousTokens(dir, tagType, FP_SP, 10, 5, 20, 7, splitter, outputDirName);
    lists[1] = ListCreator.getPreviousTokens(dir, tagType, FP_SP, 3, 5, 20, 1, splitter, outputDirName);
    lists[2] = ListCreator.getPreviousTokens(dir, tagType, FP_SP, 10, 3, 4, 7, splitter, outputDirName);
    lists[3] = ListCreator.getPreviousTokens(dir, tagType, FP_SP, 3, 3, 4, 1, splitter, outputDirName);
    lists[4] = ListCreator.getPreviousTokens(dir, tagType, FP_SP, 10, 1, 2, 7, splitter, outputDirName);
    lists[5] = ListCreator.getPreviousTokens(dir, tagType, FP_SP, 3, 1, 2, 1, splitter, outputDirName);
    listManager = new ListManager(outputDirName + "/PersonSafeList.txt", lists, outputDirName + "/greekNames.txt");

    return listManager;
  } // initListManager

  public static ListManager initListManager(String FP_SP, String path) {
    ListManager listManager;
    String lists[] = new String[6];
    if (FP_SP.compareTo("FP") == 0) {
      lists[0] = path + "/FPPERSON,lessOcc=10,minWordLen=5,maxWordLen=20,numPrevWords=7";
      lists[1] = path + "/FPPERSON,lessOcc=3,minWordLen=5,maxWordLen=20,numPrevWords=1";
      lists[2] = path + "/FPPERSON,lessOcc=10,minWordLen=3,maxWordLen=4,numPrevWords=7";
      lists[3] = path + "/FPPERSON,lessOcc=3,minWordLen=3,maxWordLen=4,numPrevWords=1";
      lists[4] = path + "/FPPERSON,lessOcc=10,minWordLen=1,maxWordLen=2,numPrevWords=7";
      lists[5] = path + "/FPPERSON,lessOcc=3,minWordLen=1,maxWordLen=2,numPrevWords=1";
    } // if
    else {
      lists[0] = path + "/SPPERSON,lessOcc=10,minWordLen=5,maxWordLen=20,numPrevWords=7";
      lists[1] = path + "/SPPERSON,lessOcc=3,minWordLen=5,maxWordLen=20,numPrevWords=1";
      lists[2] = path + "/SPPERSON,lessOcc=10,minWordLen=3,maxWordLen=4,numPrevWords=7";
      lists[3] = path + "/SPPERSON,lessOcc=3,minWordLen=3,maxWordLen=4,numPrevWords=1";
      lists[4] = path + "/SPPERSON,lessOcc=10,minWordLen=1,maxWordLen=2,numPrevWords=7";
      lists[5] = path + "/SPPERSON,lessOcc=3,minWordLen=1,maxWordLen=2,numPrevWords=1";
    } // else
    listManager = new ListManager(path + "/PersonSafeList.txt", lists, path + "/greekNames.txt");

    return listManager;
  } // initListManager

} // end of class ListManager
