/**
 * Greek Named Entity Recognizer.
 * Copyright (C) 2005  G. Lucarelli
 *
 * This file is part of Greek Named Entity Recognizer.
 *
 * Sentence Splitter is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Sentence Splitter is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Foobar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package ipl.nerc.regex.timex;

import java.util.regex.Pattern;
import java.util.regex.Matcher;

import ipl.nerc.regex.*;
import ipl.core.Tokenizer;

/**
 *             :
 *  -    :
 *       protected final String  = "";
 *  -    :
 *       private final Pattern ;
 *  -    :
 *       protected boolean match(String expr)
 *  -     match      
 *      .
 *  -     generalization   TimexFinder 
 *           .
 */


/**
 * <p>Description: Determines the categories of timex patterns.</p>
 *
 * <p>Copyright (c) 2005</p>
 * @version 0.1
 */
public class TimexRegex extends Regex {

  protected final String SEPARATOR = "SEPARATOR";
  protected final String DAY = "DAY";
  protected final String MONTH = "MONTH";
  protected final String PART_OF_THE_DAY = "PART_OF_THE_DAY";
  protected final String SEASON = "SEASON";
  protected final String ARTICLE = "ARTICLE";
  protected final String SPECIAL_DAY = "SPECIAL_DAY";
  protected final String SPELL = "SPELL"; //   ()
  protected final String ORDINAL_NUMBER = "ORDINAL_NUMBER";
  protected final String ENDING = "ENDING";
  protected final String ABBREVIATION = "ABBREVIATION";
  protected final String ATTRIBUTIVE_ADJECTIVE = "ATTRIBUTIVE_ADJECTIVE"; //  
  protected final String RELATIVE = "RELATIVE";

  private final Pattern separator;
  private final Pattern day;
  private final Pattern month;
  private final Pattern partOfTheDay;
  private final Pattern season;
  private final Pattern number;
  private final Pattern article;
  private final Pattern special_day;
  private final Pattern spell;
  private final Pattern ordinal_number;
  private final Pattern ending;
  private final Pattern abbreviation;
  private final Pattern attributive_adjective;
  private final Pattern relative;

  public TimexRegex() {
    super();

    separator = Pattern.compile("[./]", Pattern.CASE_INSENSITIVE | Pattern.UNICODE_CASE | Pattern.CANON_EQ);

    String d = "((?)|(?)|(?)|" +
               "(?)|(?)|(?)|" +
               "(?)|(?)|(?)|" +
               "(?)|(?)|(?)|" +
               "(?)|(?)|(?)|" +
               "()|()|(?)|(?)|" +
               "(?)|(?)|(?))";
    day = Pattern.compile(d);

    String m = "(()|(?)|((|)?)|(?)|(?)|" +
               "()|(?)|((|)?)|(?)|(?)|" +
               "()|(?)|((|)?)|(?)|(?)|" +
               "()|(?)|((|)?)|(?)|(?)|" +
               "()|(?)|()|(?)|(?)|(?)|(?)|" +
               "()|(?)|((|)?)|(?)|(?)|" +
               "()|(?)|((|)?)|(?)|(?)|" +
               "()|(?)|((|)?)|" +
               "()|(?)|((|)?)|(?)|(?)|" +
               "()|(?)|((|)?)|(?)|(?)|" +
               "()|(?)|((|)?)|(?)|(?)|" +
               "()|(?)|((|)?)|(?)|(?))";
    month = Pattern.compile(m);

    String pod = "(((?|)((?|))?)|((?|))|((?|))|((?|))|" +
                 "((?|))|((?|)(?|)()?)|((?|)))";
    partOfTheDay = Pattern.compile(pod, Pattern.CASE_INSENSITIVE | Pattern.UNICODE_CASE | Pattern.CANON_EQ);

    String s = "(((?|)?)|((?|)?)|((?|)((?|))?)|((?|)(?|)?))";
    season = Pattern.compile(s, Pattern.CASE_INSENSITIVE | Pattern.UNICODE_CASE | Pattern.CANON_EQ);

    number = Pattern.compile("[0-9]{1,4}", Pattern.CASE_INSENSITIVE | Pattern.UNICODE_CASE | Pattern.CANON_EQ);

    String art = "(()|()|()|()|()|()|()|()|()|()|()|()|()|())";
    article = Pattern.compile(art, Pattern.CASE_INSENSITIVE | Pattern.UNICODE_CASE | Pattern.CANON_EQ);

    String sd = "((?)|(?)|" +
                "()|()|((|))|" +
                "()|()|" +
                "(?)|(?)|" +
                "(?)|(?)|" +
                "((?|))|((?|))|" +
                "((|)?)|((|)?))";
    special_day = Pattern.compile(sd);

    String sp = "(((?|)(?||))|((?|)(?||((?|))))|" +
               "((?|)(?||((?|))))|((?|)(||||))|" +
               "((?|)(?||((?|))))|(?(?|)(?||((?|))))|" +
               "((?|)(?|)(?||))|((?|)(?|)(?||))|" +
               "((?|)(?|)(?||))|((?|)(?|)(?||))|" +
               "((?|)(?|)(?||))|((?|)?)|" +
               "((?|)?)|((?|)?)|((?|)?)|" +
               "((?|)?)|(?(?|)((?)|()|())))";
    spell = Pattern.compile(sp, Pattern.CASE_INSENSITIVE | Pattern.UNICODE_CASE | Pattern.CANON_EQ);

    String on = "(((?|)(?|?|?||))|((?|)(?|)(|?|?|))|" +
                 "((?|)(|?|?|))|((?|)(?|)(|?|?|))|" +
                 "((?|)(|?|?|))|((?|)(|?|?|))|" +
                 "((?|)(?|)(|?|?|))|((?|)(?|)(|?|?|))|" +
                 "((?|)?(?|)(|?|?|))|((?|)(?|)(|?|?|))|" +
                 "((?|)(?|)(|?|?|))|((?|)(?|)(|?|?|)))";
    ordinal_number = Pattern.compile(on, Pattern.CASE_INSENSITIVE | Pattern.UNICODE_CASE | Pattern.CANON_EQ);

    String e = "(((?|))|()|((?|)?)|(?)|(?)|(?))";
    ending = Pattern.compile(e, Pattern.CASE_INSENSITIVE | Pattern.UNICODE_CASE | Pattern.CANON_EQ);

    String abb = "(||||||)";
    abbreviation = Pattern.compile(abb, Pattern.CASE_INSENSITIVE | Pattern.UNICODE_CASE | Pattern.CANON_EQ);

    String aa = "(((?||?|)?)|((|)?)|((?|)?)|(?))";
    attributive_adjective = Pattern.compile(aa, Pattern.CASE_INSENSITIVE | Pattern.UNICODE_CASE | Pattern.CANON_EQ);

    String r = "(((?|)(|))|((?|)?)|((?|))|((?|))|" +
               "((?|))|((?|))|((?|))|((?|)(|?|?))|" +
               "((?|)(|?|?))|((?|)(|?|?))|" +
               "((?|)(|?|?))|((((?||?|)?)|((?|)))))";
    relative = Pattern.compile(r, Pattern.CASE_INSENSITIVE | Pattern.UNICODE_CASE | Pattern.CANON_EQ);
  } // constructor: TimexRegex

  public int match(Tokenizer tokenizer, int start) {
    int tokSize = tokenizer.size();
    int size = regexList.size();
    for (int i = 0; i < size; i++) {
      String expression[] = regexList.get(i).get_expression();
      int length = expression.length;
      int j = 0;
      boolean stop = false;
      while ((j < length) && !stop) {
        if (start + j >= tokSize) {
          stop = true;
        } // if
        // SEPARATOR
        else if (expression[j].compareTo(SEPARATOR) == 0) {
          if (!matchSeparator(tokenizer.get(start + j).getToken())) {
            stop = true;
          } // if
          else {
            j++;
          } // else
        } // else if
        // NUMBER
        else if (expression[j].contains("[0-9]")) {
          if (!Pattern.matches(expression[j], tokenizer.get(start + j).getToken())) {
            stop = true;
          } // if
          else {
            j++;
          } // else
        } // else if
        // ARTICLE
        else if (expression[j].compareTo(ARTICLE) == 0) {
          if (!matchArticle(tokenizer.get(start + j).getToken())) {
            stop = true;
          } // if
          else {
            j++;
          } // else
        } // else if
        // DAY
        else if (expression[j].compareTo(DAY) == 0) {
          if (!matchDay(tokenizer.get(start + j).getToken())) {
            stop = true;
          } // if
          else {
            j++;
          } // else
        } // else if
        // MONTH
        else if (expression[j].compareTo(MONTH) == 0) {
          if (!matchMonth(tokenizer.get(start + j).getToken())) {
            stop = true;
          } // if
          else {
            j++;
          } // else
        } // else if
        // PART OF THE DAY
        else if (expression[j].compareTo(PART_OF_THE_DAY) == 0) {
          if (!matchPartOfTheDay(tokenizer.get(start + j).getToken())) {
            stop = true;
          } // if
          else {
            j++;
          } // else
        } // else if
        // SEASON
        else if (expression[j].compareTo(SEASON) == 0) {
          if (!matchSeason(tokenizer.get(start + j).getToken())) {
            stop = true;
          } // if
          else {
            j++;
          } // else
        } // else if
        // SPECIAL DAY
        else if (expression[j].compareTo(SPECIAL_DAY) == 0) {
          if (!matchSpecialDay(tokenizer.get(start + j).getToken())) {
            stop = true;
          } // if
          else {
            j++;
          } // else
        } // else if
        // SPELL
        else if (expression[j].compareTo(SPELL) == 0) {
          if (!matchSpell(tokenizer.get(start + j).getToken())) {
            stop = true;
          } // if
          else {
            j++;
          } // else
        } // else if
        // ORDINAL NUMBER
        else if (expression[j].compareTo(ORDINAL_NUMBER) == 0) {
          if (!matchOrdinalNumber(tokenizer.get(start + j).getToken())) {
            stop = true;
          } // if
          else {
            j++;
          } // else
        } // else if
        // ENDING
        else if (expression[j].compareTo(ENDING) == 0) {
          if (!matchEnding(tokenizer.get(start + j).getToken())) {
            stop = true;
          } // if
          else {
            j++;
          } // else
        } // else if
        // ABBREVIATION
        else if (expression[j].compareTo(ABBREVIATION) == 0) {
          if (!matchAbbreviation(tokenizer.get(start + j).getToken())) {
            stop = true;
          } // if
          else {
            j++;
          } // else
        } // else if
        // ATTRIBUTIVE ADJECTIVE
        else if (expression[j].compareTo(ATTRIBUTIVE_ADJECTIVE) == 0) {
          if (!matchAttributiveAdjective(tokenizer.get(start + j).getToken())) {
            stop = true;
          } // if
          else {
            j++;
          } // else
        } // else if
        // RELATIVE
        else if (expression[j].compareTo(RELATIVE) == 0) {
          if (!matchRelative(tokenizer.get(start + j).getToken())) {
            stop = true;
          } // if
          else {
            j++;
          } // else
        } // else if
        /**********************************************************************/
        //         
        //        .
        /**********************************************************************/
        else {
          if (!Pattern.matches(expression[j], tokenizer.get(start + j).getToken())) {
            stop = true;
          } // if
          else {
            j++;
          } // else
        } // else
      } // while
      if (!stop && (j == length)) {
        return start + length - 1;
      } // if
    } // for
    return -1;
  } // match

  protected boolean matchSeparator(String expr) {
    Matcher matcher = separator.matcher(expr);
    if (matcher.matches()) {
      return true;
    } // if
    else {
      return false;
    } // else
  } // matchSeparator

  protected boolean matchDay(String expr) {
    Matcher matcher = day.matcher(expr);
    if (matcher.matches()) {
      return true;
    } // if
    else {
      return false;
    } // else
  } // matchDay

  protected boolean matchMonth(String expr) {
    Matcher matcher = month.matcher(expr);
    if (matcher.matches()) {
      return true;
    } // if
    else {
      return false;
    } // else
  } // matchMonth

  protected boolean matchPartOfTheDay(String expr) {
  Matcher matcher = partOfTheDay.matcher(expr);
  if (matcher.matches()) {
    return true;
  } // if
  else {
    return false;
  } // else
} // matchPartOfTheDay

  protected boolean matchSeason(String expr) {
    Matcher matcher = season.matcher(expr);
    if (matcher.matches()) {
      return true;
    } // if
    else {
      return false;
    } // else
  } // matchSeason

  protected boolean matchNumber(String expr) {
    Matcher matcher = number.matcher(expr);
    if (matcher.matches()) {
      return true;
    } // if
    else {
      return false;
    } // else
  } // matchNumber

  protected boolean matchArticle(String expr) {
    Matcher matcher = article.matcher(expr);
    if (matcher.matches()) {
      return true;
    } // if
    else {
      return false;
    } // else
  } // matchArticle

  protected boolean matchSpecialDay(String expr) {
    Matcher matcher = special_day.matcher(expr);
    if (matcher.matches()) {
      return true;
    } // if
    else {
      return false;
    } // else
  } // matchSpecialDay

  protected boolean matchSpell(String expr) {
    Matcher matcher = spell.matcher(expr);
    if (matcher.matches()) {
      return true;
    } // if
    else {
      return false;
    } // else
  } // matchSpell

  protected boolean matchOrdinalNumber(String expr) {
    Matcher matcher = ordinal_number.matcher(expr);
    if (matcher.matches()) {
      return true;
    } // if
    else {
      return false;
    } // else
  } // matchOrdinalNumber

  protected boolean matchEnding(String expr) {
    Matcher matcher = ending.matcher(expr);
    if (matcher.matches()) {
      return true;
    } // if
    else {
      return false;
    } // else
  } // matchEnding

  protected boolean matchAbbreviation(String expr) {
    Matcher matcher = abbreviation.matcher(expr);
    if (matcher.matches()) {
      return true;
    } // if
    else {
      return false;
    } // else
  } // matchAbbreviation

  protected boolean matchAttributiveAdjective(String expr) {
    Matcher matcher = attributive_adjective.matcher(expr);
    if (matcher.matches()) {
      return true;
    } // if
    else {
      return false;
    } // else
  } // matchAttributiveAdjective

  protected boolean matchRelative(String expr) {
    Matcher matcher = relative.matcher(expr);
    if (matcher.matches()) {
      return true;
    } // if
    else {
      return false;
    } // else
  } // matchRelative

  public void sort() {
    if (!regexList.isEmpty()) {
      int length = regexList.size();
      for (int j = length - 1; j >= 0; j--) {
        for (int i = 0; i < j; i++) {
          if (regexList.get(i).compareTo(regexList.get(i + 1)) < 0) {
            PatternNode temp = regexList.remove(i);
            regexList.add(i + 1, temp);
          } // if
        } // for
      } // for
    } // if
  } // sort

} // end of class TimexRegex
