/**
 * Greek Named Entity Recognizer.
 * Copyright (C) 2005  G. Lucarelli
 *
 * This file is part of Greek Named Entity Recognizer.
 *
 * Sentence Splitter is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Sentence Splitter is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Foobar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package ipl.splitter;

import java.util.regex.*;

import ipl.core.Tokenizer;

/**
 * <p>Description: Manages the feature vectors of a text.</p>
 *
 * <p>Copyright: Copyright (c) 2005</p>
 * @version 0.1
 */
public class SplitterFeaturesManager {

  private Tokenizer tokens;
  private int offset;

  public SplitterFeaturesManager(String text) {
    this.offset = 0;
    this.tokens = new Tokenizer(text, null, null);
  } // contructor: SplitterFeaturesManager

  public SplitterFeaturesManager(Tokenizer tokenizer) {
    this.offset = 0;
    this.tokens = tokenizer;
  } // contructor: SplitterFeaturesManager

  public Tokenizer getTokens() {
    return tokens;
  } // getTokens

  public int getOffset() {
    return offset;
  } // getOffset

  public void reset() {
    offset = 0;
  } // reset

  //////////////////////////////////////////////////////////////////////////////
  //           The following functions help to features retrieval.            //
  //////////////////////////////////////////////////////////////////////////////

  private int findPrevious(int from) {
    int i = from;
    while ((i >= 0) && (tokens.get(i).getToken().compareTo(".") != 0)) {
      i--;
    } // while
    return i;
  } // findPrevious

  private int findNext(int from) {
    int i = from;
    while ((i < tokens.size()) && (tokens.get(i).getToken().compareTo(".") != 0)) {
      i++;
    } // while
    return i;
  } // findNext

  private String endsToVowel(int pos) {
    if (pos < 0) {
      return "-1";
    } // if
    if (pos >= tokens.size()) {
      return "1";
    } // if

    char ch = tokens.get(pos).getToken().charAt(tokens.get(pos).getToken().length() - 1) ;
    if ( (ch == '') || (ch == '') || (ch == '') || (ch == '') || (ch == '') ||
        (ch == '') || (ch == '') ||
        (ch == '') || (ch == '') || (ch == '') || (ch == '') || (ch == '') ||
        (ch == '') || (ch == '')) {
      return "1";
    } // if
    else {
      return "-1";
    } // else
  } // endsToVowel

  private String endsToGreekN(int pos) {
    if (pos < 0) {
      return "-1";
    } // if
    if (pos >= tokens.size()) {
      return "1";
    } // if

    char ch = tokens.get(pos).getToken().charAt(tokens.get(pos).getToken().length() - 1) ;
    if (ch == '') {
      return "1";
    } // if
    else {
      return "-1";
    } // else
  } // endsToGreekN

  private String endsToGreekFinalS(int pos) {
    if (pos < 0) {
      return "-1";
    } // if
    if (pos >= tokens.size()) {
      return "1";
    } // if

    char ch = tokens.get(pos).getToken().charAt(tokens.get(pos).getToken().length() - 1) ;
    if (ch == '') {
      return "1";
    } // if
    else {
      return "-1";
    } // else
  } // endsToGreekFinalS

  private String distanceFromPreviousDot() {
    double dist = offset - findPrevious(offset - 1);

    if (dist >= 15) {
      return "1";
    } // if
    else {
      return Double.toString(((dist / 15) * 2) - 1);
    } // else
  } // distanceFromPreviousDot

  private String distanceFromNextDot() {
    double dist = findNext(offset + 1) - offset;

    if (dist >= 15) {
      return "1";
    } // if
    else {
      return Double.toString(((dist / 15) * 2) - 1);
    } // else
  } // distanceFromNextDot

  private String isGreekWord(int pos) {
    if (pos < 0) {
      return "-1";
    } // if
    if (pos >= this.tokens.size()) {
      return "1";
    } // if

    String grWord = "\\p{InGreek}*";
    String tok = tokens.get(pos).getToken();

    if (Pattern.matches(grWord, tok)) {
      return "1";
    } // if
    else {
      return "-1";
    } // else
  } // isGreekWord

  private String isEnglishWord(int pos) {
    if (pos < 0) {
      return "-1";
    } // if
    if (pos >= this.tokens.size()) {
      return "1";
    } // if

    String enWord = "[a-zA-Z]*";
    String tok = tokens.get(pos).getToken();

    if (Pattern.matches(enWord, tok)) {
      return "1";
    } // if
    else {
      return "-1";
    } // else
  } // isEnglishWord

  private String isNumber(int pos) {
    if (pos < 0) {
      return "-1";
    } // if
    if (pos >= this.tokens.size()) {
      return "1";
    } // if

    String number = "[0-9]*";
    String tok = tokens.get(pos).getToken();

    if (Pattern.matches(number, tok)) {
      return "1";
    } // if
    else {
      return "-1";
    } // else
  } // isNumber

  private String length(int pos) {
    if (pos < 0) {
      return "-1";
    } // if
    if (pos >= this.tokens.size()) {
      return "1";
    } // if

    String tok = tokens.get(pos).getToken();
    double l = tok.length();

    if (l >= 10) {
      return "1";
    } // if
    else {
      return Double.toString(((l / 10) * 2) - 1);
    } // else
  } // length

  private String isFirstCapital(int pos) {
    if (pos < 0) {
      return "-1";
    } // if
    if (pos >= this.tokens.size()) {
      return "1";
    } // if

    String tok = tokens.get(pos).getToken();

    if (tok.substring(0, 1).compareTo(tok.substring(0, 1).toUpperCase()) == 0) {
      return "1";
    } // if
    else {
      return "-1";
    } // else
  } // isFirstCapital

  private String isCapitalized(int pos) {
    if (pos < 0) {
      return "-1";
    } // if
    if (pos >= this.tokens.size()) {
      return "1";
    } // if

    String tok = tokens.get(pos).getToken();

    if (tok.compareTo(tok.toUpperCase()) == 0) {
      return "1";
    } // if
    else {
      return "-1";
    } // else
  } // isCapitalized

  private String getCategory() {
    if (!tokens.get(offset).getEndOfPeriod()) {
      return "-1";
    } // if
    else {
      return "1";
    } // else
  } // getCategory

  //////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////

  public String[] getNextFeaturesVector() {
    if ((offset = findNext(offset + 1)) >= tokens.size()) {
      return null;
    } // if

    String gw, ew;

    String featuresVector[] = new String[18];
    int featuresNum = 0;
    featuresVector[featuresNum++] = getCategory();

    featuresVector[featuresNum++] = endsToVowel(offset - 1);
    featuresVector[featuresNum++] = endsToGreekN(offset - 1);
    featuresVector[featuresNum++] = endsToGreekFinalS(offset - 1);
    featuresVector[featuresNum++] = distanceFromPreviousDot();
    featuresVector[featuresNum++] = distanceFromNextDot();

    // -1
    gw = isGreekWord(offset - 1);
    featuresVector[featuresNum++] = gw;
    ew = isEnglishWord(offset - 1);
    featuresVector[featuresNum++] = ew;
    featuresVector[featuresNum++] = isNumber(offset - 1);
    featuresVector[featuresNum++] = length(offset - 1);
    if ((gw.compareTo("1") == 0) || (ew.compareTo("1") == 0)) {
      featuresVector[featuresNum++] = isFirstCapital(offset - 1);
      featuresVector[featuresNum++] = isCapitalized(offset - 1);
    } // if
    else {
      featuresVector[featuresNum++] = "-1";
      featuresVector[featuresNum++] = "-1";
    } // else

    // +1
    gw = isGreekWord(offset + 1);
    featuresVector[featuresNum++] = gw;
    ew = isEnglishWord(offset + 1);
    featuresVector[featuresNum++] = ew;
    featuresVector[featuresNum++] = isNumber(offset + 1);
    featuresVector[featuresNum++] = length(offset + 1);
    if ((gw.compareTo("1") == 0) || (ew.compareTo("1") == 0)) {
      featuresVector[featuresNum++] = isFirstCapital(offset + 1);
      featuresVector[featuresNum++] = isCapitalized(offset + 1);
    } // if
    else {
      featuresVector[featuresNum++] = "-1";
      featuresVector[featuresNum++] = "-1";
    } // else

    return featuresVector;
  } // getNextFeaturesVector

} // end of class SplitterFeaturesManager
