/**
 * Greek Named Entity Recognizer.
 * Copyright (C) 2005  G. Lucarelli
 *
 * This file is part of Greek Named Entity Recognizer.
 *
 * Sentence Splitter is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Sentence Splitter is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Foobar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package ipl.splitter;

import java.io.*;

import ipl.util.*;
import ipl.core.Tokenizer;

/**
 * <p><b>We use libSVM package.</b><br>
 * Please read the LIBSVM-COPYRIGHT file.<br>
 * </p>
 * <p>Description: Sets < sentence > and < / sentence > tags to the
 * documents of a folder, according to a classifier.</p>
 *
 * <p>Copyright: Copyright (c) 2005</p>
 *
 * @version 0.1
 * @see <a href="http://www.csie.ntu.edu.tw/~cjlin/libsvm">libSVM</a>
 */
public class SplitterTagger {

  SentenceSplitter splitter;

  public SplitterTagger(String classifierFilename) {
    if (!(new File(classifierFilename)).isFile()) {
      System.err.println("\nError while loading the classifier!");
      System.exit(1);
    } // if
    splitter = new SentenceSplitter(classifierFilename);
  } // constructor: SplitterTagger

  public void markFiles(String inputdir) {
    if (!(new File(inputdir)).isDirectory()) {
      System.err.println("\nNot valid directory path");
      System.exit(1);
    } // if

    DirectoryManager directory = new DirectoryManager(inputdir);

    File outdir = new File(inputdir + File.separator + "annotated_data");
    if (outdir.exists()) {
      System.err.println("\nThe directory path " + outdir.getName() + " already exists");
      System.exit(1);
    } // if
    outdir.mkdir();

    File curr;

    while ((curr = directory.getNextFile()) != null) {
      String text = FileManager.read(curr.getAbsolutePath());
      Tokenizer tokenizer = new Tokenizer(text, null, null);
      splitter.setTokenizer(tokenizer);
      splitter.setSentenceTags();
      FileManager.write(tokenizer, outdir.getAbsolutePath() + File.separator + curr.getName());
    } // while
  } // mark

  private static void printParametersAndExit(String str) {
    System.err.println(str);
    System.err.println();
    System.err.println();
    System.err.println("java ipl.splitter.SplitterTagger -d <input dir path> -c <classifier>");
    System.err.println();
    System.err.println("-d");
    System.err.println("\tThe directory containing text files that you want to process");
    System.err.println("-c");
    System.err.println("\tThe classifier filename");
    System.err.println();
    System.exit(1);
  } // printParametersAndExit

  public static void main(String args[]) {
    if ((args.length < 4) || (args.length % 2 == 1)) {
      printParametersAndExit("Invalid parameters number!");
    } // if

    String c = "";
    String d = "";
    for (int i = 0; i < args.length; i += 2) {
      if (args[i].length() != 2) {
        printParametersAndExit("Invalid parameters!");
      } // if
      if (args[i].charAt(0) != '-') {
        printParametersAndExit("Invalid parameters!");
      } // if
      char param = args[i].charAt(1);
      switch (param) {
        case 'c':
          c = args[i + 1];
          if (!(new File(c)).isFile()){
            printParametersAndExit("Invalid classifier filename!");
          } // if
          break;
        case 'd':
          d = args[i + 1];
          if (!(new File(d)).isDirectory()) {
            printParametersAndExit("Invalid directory path!");
          } // if
          break;
        default:
          printParametersAndExit("Invalid parameters!");
          break;
      } // switch
    } // for

    System.out.println("please wait... ");
    SplitterTagger tagger = new SplitterTagger(c);
    tagger.markFiles(d);
  } // end of main

} // end of class SplitterTagger
